from astropy import cosmology
from scipy.optimize import fsolve
from autolens.model.profiles import mass_profiles as mp

from astropy import constants

import numpy as np

def convert_to_lens_unit(m_nfw, c_nfw, z_nfw, z_source, cosmo=cosmology.LambdaCDM):
    '''
    Trying to convert los subhalo with a given mass and concentration
    to lens units.

    m_nfw: m_200 [solar mass]
    c_nfw: concentration
    z_nfw: Redshift of the halo
    z_source: Redshift of the background source

    return kappa_s & rs [arcsec]
    '''
    H_l = (cosmo.H0).value  # Hubble constant
    G_l = 4.302e-9  # Mpc M_sun^-1 (km/s)
    J_c = 3. * H_l ** 2. / (8. * np.pi * G_l) / 10. ** (9.)  # M_sun/kpc^(3)
    C_v = 3 * 10 ** 5.  # light velocity [km/s]
    G_crit = 4.302e-6  # kpc M_sun^-1 (km/s)^2

    r200 = (m_nfw / (200. * J_c * (4. * np.pi / 3.))) ** (1. / 3.)  # R_200
    c = c_nfw
    # print(c)
    de_c = 200. / 3. * (c * c * c / (np.log(1. + c) - c / (1. + c)))  # delta_c
    rs = r200 / c  # rs
    rhos = J_c * de_c  # rhos
    s_crit = (C_v * C_v * cosmo.angular_diameter_distance(z_source) / (4. * np.pi *
                                                                       cosmo.angular_diameter_distance(z_nfw) *
                                                                       cosmo.angular_diameter_distance_z1z2(z_nfw,
                                                                                                            z_source)) / (
                          1000.0
                          * G_crit)).value  # critical density at los halo
    kappa_s = rs * rhos / s_crit
    arctokpc = np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) *
                                         1000.0).value
    rs_arcsec = rs / arctokpc
    return kappa_s, rs_arcsec

def convert_to_physical_unit(kappa_s, rs_arcsec, z_nfw, z_source, cosmo=cosmology.LambdaCDM):
    '''
    Trying to convert NFW lens units to physical units.

    kappa_s: rhos*rs/s_crit
    rs_arcsec: scale radius [arcsec]
    z_nfw: Redshift of the halo
    z_source: Redshift of the background source

    return m200 [solar mass] & concentration: c
    '''
    H_l = cosmo.H0.value  # Hubble constant
    G_l = 4.302e-9  # Mpc M_sun^-1 (km/s) # Gravitational constnat
    J_c = 3. * H_l ** 2. / (8. * np.pi * G_l) / 10. ** (9.)  # cosmic average density M_sun/kpc^(3)
    C_v = 3 * 10 ** 5.  # light velocity [km/s]
    G_crit = 4.302e-6  # kpc M_sun^-1 (km/s)^2 # Gravitational constant

    arctokpc = np.pi / 180.0 / 3600.0 * (cosmo.angular_diameter_distance(z_nfw) *
                                         1000.0).value
    rs = rs_arcsec * arctokpc

    s_crit = (C_v * C_v * cosmo.angular_diameter_distance(z_source) / (4. * np.pi *
                                                                       cosmo.angular_diameter_distance(z_nfw) *
                                                                       cosmo.angular_diameter_distance_z1z2(z_nfw,
                                                                                                            z_source)) / (
                          1000.0
                          * G_crit)).value  # critical density

    rhos = kappa_s * s_crit / rs  # rho_s

    print(rhos)

    de_c = rhos / J_c  # delta_c

    c = fsolve(solve_c, 10.0, args=(de_c,))[0]  # concentration

    r200 = c * rs  # R_200

    m200 = 200.0 * (4. / 3. * np.pi) * J_c * r200 ** 3.0  # M_200

    return m200, c

def solve_c(c, de_c):
    '''
    Equation need for solving concentration c for a given delta_c
    '''
    return 200.0/3.0*(c*c*c/(np.log(1+c) - c/(1+c))) - de_c

def convert_to_physical_unit_tNFW(kappa_s, rs_arcsec, z_nfw, z_source, cosmo=cosmology.LambdaCDM, rt_arcsec=None):
    '''
    Trying to convert tNFW lens units to physical units.

    kappa_s: rhos*rs/s_crit

    rs_arcsec: scale radius [arcsec]

    rt_arcsec: truncation radius [arcsec]
    if rt_arcsec is not given, then rt is set to be 2.0*R_200

    z_nfw: Redshift of the halo

    z_source: Redshift of the background source

    return m200 [solar mass], m_t [solar_mass], & concentration: c, r200 [kpc]
    '''
    H_l = (cosmo.H0).value # Hubble constant
    G_l = 4.302e-9 # Mpc M_sun^-1 (km/s) # Gravitational constnat
    J_c = cosmo.critical_density(z=0.6).to('solMass / kpc^3').value# cosmic average density M_sun/kpc^(3)
    C_v = constants.c.to('km / s')
    G_crit = constants.G.to( 'kpc km2 / (solMass s2)')

    print('Cosmic average density = ', J_c)

    arctokpc = np.pi/180.0/3600.0*(cosmo.angular_diameter_distance(z_nfw)*
            1000.0).value

    print('arcsec to kpc = ', arctokpc)

    print('scae radius in arcsec = ', rs_arcsec)

    rs = rs_arcsec*arctokpc

    print('scale radius in kpc = ', rs)

    s_crit = (C_v*C_v*cosmo.angular_diameter_distance(z_source)/(4.*np.pi*
            cosmo.angular_diameter_distance(z_nfw)*
            cosmo.angular_diameter_distance_z1z2(z_nfw, z_source))/(1000.0
                    *G_crit)) # critical density

    print('s_crit (native units) = ', s_crit)

    s_crit = s_crit.value

    print('critical density = ', s_crit)

    rhos = kappa_s*s_crit/rs # rho_s

    print('rho = ', rhos)

    de_c = rhos/J_c # delta_c

    print('cosmic density = ', J_c)
    print('cosmic density astropy = ', cosmo.critical_density(z=0.0).to('solMass / kpc^3'))
    print('delta con = ', de_c)

    c = fsolve(solve_c, 10.0, args=(de_c,))[0] # concentration

    r200 = c*rs # R_200

    m200 = 200.0*(4./3.*np.pi)*J_c*r200**3.0 # M_200

    if rt_arcsec is not None:
        tau = rt_arcsec/rs_arcsec
    else:
        tau = 2.0*c


    # if rt is given, calculate the tau
    # if rt is not given, set rt = 2.0*r_200, which means tau = 2.0*c

    m_t = m200*(tau*tau/(tau*tau + 1.0)**2.0)*((tau*tau - 1)*np.log(tau) +
            tau*np.pi - (tau*tau + 1))

    # m_t is the truncation mass.

    return m200, m_t, c, r200

def convert_to_physical_unit_tNFW_kpc(kappa_s, rs_kpc, z_nfw, z_source, cosmo=cosmology.LambdaCDM, rt_kpc=None):
    '''
    Trying to convert tNFW lens units to physical units.

    kappa_s: rhos*rs/s_crit

    rs_arcsec: scale radius [arcsec]

    rt_arcsec: truncation radius [arcsec]
    if rt_arcsec is not given, then rt is set to be 2.0*R_200

    z_nfw: Redshift of the halo

    z_source: Redshift of the background source

    return m200 [solar mass], m_t [solar_mass], & concentration: c, r200 [kpc]
    '''
    H_l = (cosmo.H0).value # Hubble constant
    G_l = 4.302e-9 # Mpc M_sun^-1 (km/s) # Gravitational constnat
    J_c = cosmo.critical_density(z=0.6).to('solMass / kpc^3').value# cosmic average density M_sun/kpc^(3)
    C_v = constants.c.to('km / s')
    G_crit = constants.G.to( 'kpc km2 / (solMass s2)')

    print('Cosmic average density = ', J_c)

    arctokpc = np.pi/180.0/3600.0*(cosmo.angular_diameter_distance(z_nfw)*
            1000.0).value

    print('arcsec to kpc = ', arctokpc)

    rs = rs_kpc

    print('scale radius in kpc = ', rs)

    s_crit = (C_v*C_v*cosmo.angular_diameter_distance(z_source)/(4.*np.pi*
            cosmo.angular_diameter_distance(z_nfw)*
            cosmo.angular_diameter_distance_z1z2(z_nfw, z_source))/(1000.0
                    *G_crit)) # critical density

    print('s_crit (native units) = ', s_crit)

    s_crit = s_crit.value

    print('critical density = ', s_crit)

    rhos = kappa_s*s_crit/rs # rho_s

    print('rho = ', rhos)

    de_c = rhos/J_c # delta_c

    print('cosmic density = ', J_c)
    print('cosmic density astropy = ', cosmo.critical_density(z=0.0).to('solMass / kpc^3'))
    print('delta con = ', de_c)

    c = fsolve(solve_c, 10.0, args=(de_c,))[0] # concentration

    r200 = c*rs # R_200

    m200 = 200.0*(4./3.*np.pi)*J_c*r200**3.0 # M_200

    if rt_kpc is not None:
        tau = rt_kpc / rs_kpc
    else:
        tau = 2.0*c


    # if rt is given, calculate the tau
    # if rt is not given, set rt = 2.0*r_200, which means tau = 2.0*c

    m_t = m200*(tau*tau/(tau*tau + 1.0)**2.0)*((tau*tau - 1)*np.log(tau) +
            tau*np.pi - (tau*tau + 1))

    # m_t is the truncation mass.

    return m200, m_t, c, r200

def r200_from_rs_arcsec(kappa_s, rs_arcsec, cosmo=cosmology.LambdaCDM):
    '''
    Trying to convert tNFW lens units to physical units.

    kappa_s: rhos*rs/s_crit

    rs_arcsec: scale radius [arcsec]

    rt_arcsec: truncation radius [arcsec]
    if rt_arcsec is not given, then rt is set to be 2.0*R_200

    z_nfw: Redshift of the halo

    z_source: Redshift of the background source

    return m200 [solar mass], m_t [solar_mass], & concentration: c, r200 [kpc]
    '''

    z_lens = 0.6
    z_source = 2.5
    J_c = cosmo.critical_density(z=z_lens).to('solMass / kpc^3').value# cosmic average density M_sun/kpc^(3)
    C_v = constants.c.to('km / s')
    G_crit = constants.G.to( 'kpc km2 / (solMass s2)')

    print('Cosmic average density = ', J_c)

    arctokpc = np.pi/180.0/3600.0*(cosmo.angular_diameter_distance(z_nfw)*1000.0).value

    rs = rs_arcsec*arctokpc

    s_crit = (C_v*C_v*cosmo.angular_diameter_distance(z_source)/(4.*np.pi*
            cosmo.angular_diameter_distance(z_lens)*
            cosmo.angular_diameter_distance_z1z2(z_nfw, z_source))/(1000.0
                    *G_crit)) # critical density

    print(J_c)
    print(s_crit)

    s_crit = s_crit.value
    rhos = kappa_s*s_crit/rs # rho_s
    de_c = rhos/J_c # delta_c
    c = fsolve(solve_c, 10.0, args=(de_c,))[0] # concentration
    r200 = c*rs # R_200
    print(r200)
    stop
    r200 = r200 / arctokpc
    print(arctokpc)
    print(r200)
    stop

    return r200

z_source = 2.5
z_nfw = 0.6

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(kappa_s=0.5, rs_arcsec=5.0, z_nfw=0.6, z_source=2.5,
                                                                   rt_arcsec=10.0,
                                                                   cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3))

print(mass_200, m_t, concentration, r200)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW_kpc(kappa_s=0.5,z_nfw=0.6, z_source=2.5,
                                                                   rs_kpc=3.0, rt_kpc=7.0,
                                                                   cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3))

print(mass_200, m_t, concentration, r200*6.68549148608755)


r200 = r200_from_rs_arcsec(kappa_s=2.0, rs_arcsec=10.0, cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3))
print(r200)
stop

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(kappa_s=0.002, rs_arcsec=30.0, z_nfw=0.6, z_source=2.5,
                                                   cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)  )

print("{:.4e}".format(mass_200), concentration)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(kappa_s=0.00001, rs_arcsec=5.0, z_nfw=0.6, z_source=2.5,
                                                   cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)  )

print("{:.4e}".format(mass_200), concentration)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(kappa_s=0.00001, rs_arcsec=1.0, z_nfw=0.6, z_source=2.5,
                                                   cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)  )

print("{:.4e}".format(mass_200), concentration)

mass_200, m_t, concentration, r200 = convert_to_physical_unit_tNFW(kappa_s=0.00001, rs_arcsec=30.0, z_nfw=0.6, z_source=2.5,
                                                   cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)  )

print("{:.4e}".format(mass_200), concentration)


import numba

from autofit import conf

"""
Depending on if we're using a super computer, we want two different numba decorators:

If on laptop:

@numba.jit(nopython=True, cache=True, parallel=False)

If on super computer:

@numba.jit(nopython=True, cache=False, parallel=True)
"""

nopython = conf.instance.general.get("numba", "nopython", bool)
cache = conf.instance.general.get("numba", "cache", bool)
parallel = conf.instance.general.get("numba", "parallel", bool)


def jit(nopython=nopython, cache=cache, parallel=parallel):
    def wrapper(func):
        return numba.jit(func, nopython=nopython, cache=cache, parallel=parallel)

    return wrapper




import os

from autofit import conf
from autofit.optimize import non_linear as nl
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration import integration_util
from test.simulation import simulation_util

test_type = 'lens_only'
test_name = "lens_x2_galaxies_separate"

test_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + 'output/'
config_path = test_path + 'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(data_type='lens_only_x2_galaxies', data_resolution='LSST')
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data, assert_optimizer_pickle_matches=False)

def make_pipeline(test_name):
    
    def modify_mask_function(image):
        return msk.Mask.circular(shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=5.)

    class LensPlaneGalaxy0Phase(ph.LensPlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens_0.sersic.centre_0 = -1.0
            self.lens_galaxies.lens_0.sersic.centre_1 = -1.0

    phase1 = LensPlaneGalaxy0Phase(phase_name='phase_1', phase_folders=[test_type, test_name],
                                   lens_galaxies=dict(lens_0=gm.GalaxyModel(sersic=lp.EllipticalSersic)),
                                   mask_function=modify_mask_function, optimizer_class=nl.MultiNest)

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    class LensPlaneGalaxy1Phase(ph.LensPlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens_0 = results.from_phase('phase_1').\
                constant.lens_galaxies.lens_0

            self.lens_galaxies.lens_1.sersic.centre_0 = 1.0
            self.lens_galaxies.lens_1.sersic.centre_1 = 1.0

    phase2 = LensPlaneGalaxy1Phase(phase_name='phase_2', phase_folders=[test_type, test_name],
                                   lens_galaxies=dict(lens_0=gm.GalaxyModel(sersic=lp.EllipticalSersic),
                                                      lens_1=gm.GalaxyModel(sersic=lp.EllipticalSersic)),
                                   mask_function=modify_mask_function, optimizer_class=nl.MultiNest)

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    class LensPlaneBothGalaxyPhase(ph.LensPlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens_0 = results.from_phase('phase_1').\
                variable.lens_galaxies.lens_0

            self.lens_galaxies.lens_1 = results.from_phase('phase_2').\
                variable.lens_galaxies.lens_0

            self.lens_galaxies.lens_0.sersic.centre_0 = -1.0
            self.lens_galaxies.lens_0.sersic.centre_1 = -1.0
            self.lens_galaxies.lens_1.sersic.centre_0 = 1.0
            self.lens_galaxies.lens_1.sersic.centre_1 = 1.0

    phase3 = LensPlaneBothGalaxyPhase(phase_name='phase_3', phase_folders=[test_type, test_name],
                                      lens_galaxies=dict(lens_0=gm.GalaxyModel(sersic=lp.EllipticalSersic),
                                                         lens_1=gm.GalaxyModel(sersic=lp.EllipticalSersic)),
                                      mask_function=modify_mask_function, optimizer_class=nl.MultiNest)

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 60
    phase3.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(test_name, phase1, phase2, phase3)


if __name__ == "__main__":
    pipeline()



import os

from autofit import conf
from autofit.mapper import prior
from autofit.optimize import non_linear as nl
from autofit.tools import phase as autofit_ph
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from test.integration import integration_util
from test.simulation import simulation_util

test_type = 'grid_search'
test_name = "multinest_grid_subhalo"

test_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + 'output/'
config_path = test_path + 'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(data_type='no_lens_light_and_source_smooth', data_resolution='LSST')
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data)# , assert_optimizer_pickle_matches=False)


def make_pipeline(test_name):
    
    class QuickPhase(ph.LensSourcePlanePhase):

        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = prior.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
            self.lens_galaxies.lens.mass.centre_1 = prior.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
            self.lens_galaxies.lens.mass.axis_ratio = prior.UniformPrior(lower_limit=0.65, upper_limit=0.75)
            self.lens_galaxies.lens.mass.phi = prior.UniformPrior(lower_limit=40.0, upper_limit=50.0)
            self.lens_galaxies.lens.mass.einstein_radius = prior.UniformPrior(lower_limit=1.55, upper_limit=1.65)

            self.source_galaxies.source.light.centre_0 = prior.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
            self.source_galaxies.source.light.centre_1 = prior.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
            self.source_galaxies.source.light.axis_ratio = prior.UniformPrior(lower_limit=0.75, upper_limit=0.85)
            self.source_galaxies.source.light.phi = prior.UniformPrior(lower_limit=50.0, upper_limit=70.0)
            self.source_galaxies.source.light.intensity = prior.UniformPrior(lower_limit=0.35, upper_limit=0.45)
            self.source_galaxies.source.light.effective_radius = prior.UniformPrior(lower_limit=0.45, upper_limit=0.55)
            self.source_galaxies.source.light.sersic_index = prior.UniformPrior(lower_limit=0.9, upper_limit=1.1)

    phase1 = QuickPhase(phase_name='phase_1', phase_folders=[test_type, test_name],
                        lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                        source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalSersic)),
                        optimizer_class=nl.MultiNest)

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    class GridPhase(autofit_ph.as_grid_search(ph.LensSourcePlanePhase)):

        @property
        def grid_priors(self):
            return [self.variable.lens_galaxies.subhalo.mass.centre_0,
                    self.variable.lens_galaxies.subhalo.mass.centre_1]

        def pass_priors(self, results):

            ### Lens Mass, PL -> PL, Shear -> Shear ###

            # self.lens_galaxies.lens = results.from_phase('phase_1').\
            #     constant.lens_galaxies.lens

            ### Lens Subhalo, Adjust priors to physical masses (10^6 - 10^10) and concentrations (6-24)

            self.lens_galaxies.subhalo.mass.kappa_s = prior.UniformPrior(lower_limit=0.0005, upper_limit=0.2)
            self.lens_galaxies.subhalo.mass.scale_radius = prior.UniformPrior(lower_limit=0.001, upper_limit=1.0)
            self.lens_galaxies.subhalo.mass.centre_0 = prior.UniformPrior(lower_limit=-2.0, upper_limit=2.0)
            self.lens_galaxies.subhalo.mass.centre_1 = prior.UniformPrior(lower_limit=-2.0, upper_limit=2.0)

            ### Source Light, Sersic -> Sersic ###

            # self.source_galaxies.source.light.centre = results.from_phase('phase_1').\
            #     variable_absolute(a=0.05).source_galaxies.source.light.centre
            #
            # self.source_galaxies.source.light.intensity = results.from_phase('phase_1').\
            #     variable_relative(r=0.5).source_galaxies.source.light.intensity
            #
            # self.source_galaxies.source.light.effective_radius = results.from_phase('phase_1').\
            #     variable_relative(r=0.5).source_galaxies.source.light.effective_radius
            #
            # self.source_galaxies.source.light.sersic_index = results.from_phase('phase_1').\
            #     variable_relative(r=0.5).source_galaxies.source.light.sersic_index
            #
            # self.source_galaxies.source.light.axis_ratio = results.from_phase('phase_1').\
            #     constant.source_galaxies.source.light.axis_ratio
            #
            # self.source_galaxies.source.light.phi = results.from_phase('phase_1').\
            #     constant.source_galaxies.source.light.phi


    phase2 = GridPhase(phase_name='phase_2', phase_folders=[test_type, test_name],
                       lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal),
                                          subhalo=gm.GalaxyModel(mass=mp.SphericalTruncatedNFWChallenge)),
                       source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalSersic)),
                       optimizer_class=nl.MultiNest,
                       number_of_steps=2)

    phase2.optimizer.const_efficiency_mode = True

    return pl.PipelineImaging(test_name, phase1, phase2)


if __name__ == "__main__":
    pipeline()



import pytest

from autolens import exc, dimensions as dim

from test.unit.mock.mock_cosmology import MockCosmology


class TestLength(object):

    def test__conversions_from_arcsec_to_kpc_and_back__errors_raised_if_no_kpc_per_arcsec(self):
        unit_arcsec = dim.Length(value=2.0)

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit_length == 'arcsec'

        unit_arcsec = unit_arcsec.convert(unit_length='arcsec')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'arcsec'

        unit_kpc = unit_arcsec.convert(unit_length='kpc', kpc_per_arcsec=2.0)

        assert unit_kpc == 4.0
        assert unit_kpc.unit == 'kpc'

        unit_kpc = unit_kpc.convert(unit_length='kpc')

        assert unit_kpc == 4.0
        assert unit_kpc.unit == 'kpc'

        unit_arcsec = unit_kpc.convert(unit_length='arcsec', kpc_per_arcsec=2.0)

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'arcsec'

        with pytest.raises(exc.UnitsException):
            unit_arcsec.convert(unit_length='kpc')
            unit_kpc.convert(unit_length='arcsec')
            unit_arcsec.convert(unit_length='lol')


class TestLuminosity(object):

    def test__conversions_from_eps_and_counts_and_back__errors_raised_if_no_exposure_time(self):

        unit_eps = dim.Luminosity(value=2.0)

        assert unit_eps == 2.0
        assert unit_eps.unit_luminosity == 'eps'

        unit_eps = unit_eps.convert(unit_luminosity='eps')

        assert unit_eps == 2.0
        assert unit_eps.unit == 'eps'

        unit_counts = unit_eps.convert(unit_luminosity='counts', exposure_time=2.0)

        assert unit_counts == 4.0
        assert unit_counts.unit == 'counts'

        unit_counts = unit_counts.convert(unit_luminosity='counts')

        assert unit_counts == 4.0
        assert unit_counts.unit == 'counts'

        unit_eps = unit_counts.convert(unit_luminosity='eps', exposure_time=2.0)

        assert unit_eps == 2.0
        assert unit_eps.unit == 'eps'

        with pytest.raises(exc.UnitsException):
            unit_eps.convert(unit_luminosity='counts')
            unit_counts.convert(unit_luminosity='eps')
            unit_eps.convert(unit_luminosity='lol')


class TestMass(object):

    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_no_exposure_time(self):

        mass_angular = dim.Mass(value=2.0)

        assert mass_angular == 2.0
        assert mass_angular.unit_mass == 'angular'

        # angular -> angular, stays 2.0

        mass_angular = mass_angular.convert(unit_mass='angular')

        assert mass_angular == 2.0
        assert mass_angular.unit == 'angular'

        # angular -> solMass, converts to 2.0 * 2.0 = 4.0

        mas_sol_mass = mass_angular.convert(unit_mass='solMass',critical_surface_density=2.0)

        assert mas_sol_mass == 4.0
        assert mas_sol_mass.unit == 'solMass'

        # solMass -> solMass, stays 4.0

        mas_sol_mass = mas_sol_mass.convert(unit_mass='solMass')

        assert mas_sol_mass == 4.0
        assert mas_sol_mass.unit == 'solMass'

        # solMass -> angular, stays 4.0

        mass_angular = mas_sol_mass.convert(unit_mass='angular', critical_surface_density=2.0)

        assert mass_angular == 2.0
        assert mass_angular.unit == 'angular'

        with pytest.raises(exc.UnitsException):
            mass_angular.convert(unit_mass='solMass')
            mas_sol_mass.convert(unit_mass='angular')
            mass_angular.convert(unit_mass='lol')


class TestMassOverLuminosity(object):

    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_critical_mass_density(self):

        unit_angular = dim.MassOverLuminosity(value=2.0)

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / eps'

        unit_angular = unit_angular.convert(unit_mass='angular', unit_luminosity='eps')

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / eps'

        unit_sol_mass = unit_angular.convert(unit_mass='solMass', critical_surface_density=2.0,
                                             unit_luminosity='eps')

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == 'solMass / eps'

        unit_sol_mass = unit_sol_mass.convert(unit_mass='solMass', unit_luminosity='eps')

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == 'solMass / eps'

        unit_angular = unit_sol_mass.convert(unit_mass='angular', critical_surface_density=2.0,
                                             unit_luminosity='eps')

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / eps'

        with pytest.raises(exc.UnitsException):
            unit_angular.convert(unit_mass='solMass', unit_luminosity='eps')
            unit_sol_mass.convert(unit_mass='angular', unit_luminosity='eps')
            unit_angular.convert(unit_mass='lol', unit_luminosity='eps')
            
    def test__conversions_from_eps_and_counts_and_back__errors_raised_if_no_exposure_time(self):

        unit_eps = dim.MassOverLuminosity(value=2.0)

        assert unit_eps == 2.0
        assert unit_eps.unit == 'angular / eps'

        unit_eps = unit_eps.convert(unit_mass='angular', unit_luminosity='eps')

        assert unit_eps == 2.0
        assert unit_eps.unit == 'angular / eps'

        unit_counts = unit_eps.convert(unit_mass='angular', exposure_time=2.0, unit_luminosity='counts')

        assert unit_counts == 1.0
        assert unit_counts.unit == 'angular / counts'

        unit_counts = unit_counts.convert(unit_mass='angular', unit_luminosity='counts')

        assert unit_counts == 1.0
        assert unit_counts.unit == 'angular / counts'

        unit_eps = unit_counts.convert(unit_mass='angular', exposure_time=2.0, unit_luminosity='eps')

        assert unit_eps == 2.0
        assert unit_eps.unit == 'angular / eps'

        with pytest.raises(exc.UnitsException):
            unit_eps.convert(unit_mass='angular', unit_luminosity='eps')
            unit_counts.convert(unit_mass='angular', unit_luminosity='eps')
            unit_eps.convert(unit_mass='lol', unit_luminosity='eps')


class TestMassOverLength2(object):

    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_critical_mass_density(self):

        unit_angular = dim.MassOverLength2(value=2.0)

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / arcsec^2'

        unit_angular = unit_angular.convert(unit_mass='angular', unit_length='arcsec')

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / arcsec^2'

        unit_sol_mass = unit_angular.convert(unit_mass='solMass', critical_surface_density=2.0,
                                             unit_length='arcsec')

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == 'solMass / arcsec^2'

        unit_sol_mass = unit_sol_mass.convert(unit_mass='solMass', unit_length='arcsec')

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == 'solMass / arcsec^2'

        unit_angular = unit_sol_mass.convert(unit_mass='angular', critical_surface_density=2.0,
                                             unit_length='arcsec')

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / arcsec^2'

        with pytest.raises(exc.UnitsException):
            unit_angular.convert(unit_mass='solMass', unit_length='eps')
            unit_sol_mass.convert(unit_mass='angular', unit_length='eps')
            unit_angular.convert(unit_mass='lol', unit_length='eps')

    def test__conversions_from_arcsec_to_kpc_and_back__errors_raised_if_no_kpc_per_arcsec(self):

        unit_arcsec = dim.MassOverLength2(value=2.0, unit_mass='solMass')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'solMass / arcsec^2'

        unit_arcsec = unit_arcsec.convert(unit_length='arcsec', unit_mass='solMass')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'solMass / arcsec^2'

        unit_kpc = unit_arcsec.convert(unit_length='kpc', kpc_per_arcsec=2.0, unit_mass='solMass')

        assert unit_kpc == 2.0 / 2.0**2.0
        assert unit_kpc.unit == 'solMass / kpc^2'

        unit_kpc = unit_kpc.convert(unit_length='kpc', unit_mass='solMass')

        assert unit_kpc == 2.0 / 2.0**2.0
        assert unit_kpc.unit == 'solMass / kpc^2'

        unit_arcsec = unit_kpc.convert(unit_length='arcsec', kpc_per_arcsec=2.0, unit_mass='solMass')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'solMass / arcsec^2'

        with pytest.raises(exc.UnitsException):
            unit_arcsec.convert(unit_length='kpc', unit_mass='solMass')
            unit_kpc.convert(unit_length='arcsec', unit_mass='solMass')
            unit_arcsec.convert(unit_length='lol', unit_mass='solMass')


class TestMassOverLength3(object):

    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_critical_mass_density(self):

        unit_angular = dim.MassOverLength3(value=2.0)

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / arcsec^3'

        unit_angular = unit_angular.convert(unit_mass='angular', unit_length='arcsec')

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / arcsec^3'

        unit_sol_mass = unit_angular.convert(unit_mass='solMass', critical_surface_density=2.0,
                                             unit_length='arcsec')

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == 'solMass / arcsec^3'

        unit_sol_mass = unit_sol_mass.convert(unit_mass='solMass', unit_length='arcsec')

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == 'solMass / arcsec^3'

        unit_angular = unit_sol_mass.convert(unit_mass='angular', critical_surface_density=2.0,
                                             unit_length='arcsec')

        assert unit_angular == 2.0
        assert unit_angular.unit == 'angular / arcsec^3'

        with pytest.raises(exc.UnitsException):
            unit_angular.convert(unit_mass='solMass', unit_length='eps')
            unit_sol_mass.convert(unit_mass='angular', unit_length='eps')
            unit_angular.convert(unit_mass='lol', unit_length='eps')

    def test__conversions_from_arcsec_to_kpc_and_back__errors_raised_if_no_kpc_per_arcsec(self):

        unit_arcsec = dim.MassOverLength3(value=2.0, unit_mass='solMass')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'solMass / arcsec^3'

        unit_arcsec = unit_arcsec.convert(unit_length='arcsec', unit_mass='solMass')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'solMass / arcsec^3'

        unit_kpc = unit_arcsec.convert(unit_length='kpc', kpc_per_arcsec=2.0, unit_mass='solMass')

        assert unit_kpc == 2.0 / 2.0**3.0
        assert unit_kpc.unit == 'solMass / kpc^3'

        unit_kpc = unit_kpc.convert(unit_length='kpc', unit_mass='solMass')

        assert unit_kpc == 2.0 / 2.0**3.0
        assert unit_kpc.unit == 'solMass / kpc^3'

        unit_arcsec = unit_kpc.convert(unit_length='arcsec', kpc_per_arcsec=2.0, unit_mass='solMass')

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == 'solMass / arcsec^3'

        with pytest.raises(exc.UnitsException):
            unit_arcsec.convert(unit_length='kpc', unit_mass='solMass')
            unit_kpc.convert(unit_length='arcsec', unit_mass='solMass')
            unit_arcsec.convert(unit_length='lol', unit_mass='solMass')


class MockDimensionsProfile(dim.DimensionsProfile):

    def __init__(self,
                 position: dim.Position = None,
                 param_float: float = None,
                 length: dim.Length = None,
                 luminosity : dim.Luminosity = None,
                 mass : dim.Mass = None,
                 mass_over_luminosity: dim.MassOverLuminosity = None):

        super(MockDimensionsProfile, self).__init__()

        self.position = position
        self.param_float = param_float
        self.luminosity = luminosity
        self.length = length
        self.mass = mass
        self.mass_over_luminosity = mass_over_luminosity

    @dim.convert_units_to_input_units
    def unit_length_calc(self, length_input : dim.Length, redshift_profile=None, cosmology=MockCosmology(),
                         unit_length='arcsec', **kwargs):

        return dim.Length(self.length + length_input, self.unit_length)

    @dim.convert_units_to_input_units
    def unit_luminosity_calc(self,
                             luminosity_input : dim.Luminosity = None, redshift_profile=None, cosmology=MockCosmology(),
                             unit_luminosity='eps',
                             exposure_time : float = None, **kwargs):

        return dim.Luminosity(self.luminosity + luminosity_input, self.unit_luminosity)

    @dim.convert_units_to_input_units
    def unit_mass_calc(self,
                       mass_input : dim.Mass = None, redshift_profile=None, redshift_source=1.0, cosmology=MockCosmology(),
                       unit_mass='angular', **kwargs):

        return dim.Mass(self.mass + mass_input, self.unit_mass)


class TestDimensionsProfile(object):

    class TestUnitProperties(object):

        def test__unit_length__extracted_from_profile(self):

            profile = MockDimensionsProfile(length=dim.Length(value=3.0, unit_length='arcsec'))

            assert profile.unit_length == 'arcsec'

            profile = MockDimensionsProfile(length=dim.Length(value=3.0, unit_length='kpc'))

            assert profile.unit_length == 'kpc'

            profile = MockDimensionsProfile(length=1.0)

            assert profile.unit_length == None

        def test__unit_luminosity__extracted_from_profile(self):

            profile = MockDimensionsProfile(luminosity=dim.Luminosity(value=3.0, unit_luminosity='eps'),
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_luminosity='eps'))

            assert profile.unit_luminosity == 'eps'

            profile = MockDimensionsProfile(luminosity=1.0,
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_luminosity='eps'))

            assert profile.unit_luminosity == 'eps'

            profile = MockDimensionsProfile(luminosity=dim.Luminosity(value=3.0, unit_luminosity='eps'),
                                            mass_over_luminosity=1.0)

            assert profile.unit_luminosity == 'eps'

            profile = MockDimensionsProfile(luminosity=dim.Luminosity(value=3.0, unit_luminosity='counts'),
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_luminosity='counts'))

            assert profile.unit_luminosity == 'counts'

            profile = MockDimensionsProfile(luminosity=1.0,
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_luminosity='counts'))

            assert profile.unit_luminosity == 'counts'


            profile = MockDimensionsProfile(luminosity=dim.Luminosity(value=3.0, unit_luminosity='counts'),
                                            mass_over_luminosity=1.0)

            assert profile.unit_luminosity == 'counts'

            profile = MockDimensionsProfile(luminosity=1.0, mass_over_luminosity=1.0)

            assert profile.unit_luminosity == None
            
        def test__unit_mass__extracted_from_profile(self):

            profile = MockDimensionsProfile(mass=dim.Mass(value=3.0, unit_mass='angular'),
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_mass='angular'))

            assert profile.unit_mass == 'angular'

            profile = MockDimensionsProfile(mass=1.0,
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_mass='angular'))

            assert profile.unit_mass == 'angular'

            profile = MockDimensionsProfile(mass=dim.Mass(value=3.0, unit_mass='angular'),
                                            mass_over_luminosity=1.0)

            assert profile.unit_mass == 'angular'

            profile = MockDimensionsProfile(mass=dim.Mass(value=3.0, unit_mass='solMass'),
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_mass='solMass'))

            assert profile.unit_mass == 'solMass'

            profile = MockDimensionsProfile(mass=1.0,
                                            mass_over_luminosity=dim.MassOverLuminosity(value=3.0, unit_mass='solMass'))

            assert profile.unit_mass == 'solMass'


            profile = MockDimensionsProfile(mass=dim.Mass(value=3.0, unit_mass='solMass'),
                                            mass_over_luminosity=1.0)

            assert profile.unit_mass == 'solMass'

            profile = MockDimensionsProfile(mass=1.0, mass_over_luminosity=1.0)

            assert profile.unit_mass == None


    class TestUnitConversions(object):

        def test__arcsec_to_kpc_conversions_of_length__float_and_tuple_length__conversion_converts_values(self):

            profile_arcsec = MockDimensionsProfile(
                position=(dim.Length(1.0, 'arcsec'), dim.Length(2.0, 'arcsec')),
                param_float=2.0,
                length=dim.Length(value=3.0, unit_length='arcsec'),
                luminosity=dim.Luminosity(value=4.0, unit_luminosity='eps'),
                mass=dim.Mass(value=5.0, unit_mass='angular'),
                mass_over_luminosity=dim.MassOverLuminosity(value=6.0, unit_luminosity='eps', unit_mass='angular'))

            assert profile_arcsec.position == (1.0, 2.0)
            assert profile_arcsec.position[0].unit_length == 'arcsec'
            assert profile_arcsec.position[1].unit_length == 'arcsec'
            assert profile_arcsec.param_float == 2.0
            assert profile_arcsec.length == 3.0
            assert profile_arcsec.length.unit_length == 'arcsec'
            assert profile_arcsec.luminosity == 4.0
            assert profile_arcsec.luminosity.unit_luminosity == 'eps'
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == 'angular'
            assert profile_arcsec.mass_over_luminosity == 6.0
            assert profile_arcsec.mass_over_luminosity.unit == 'angular / eps'

            profile_arcsec = profile_arcsec.new_profile_with_units_converted(unit_length='arcsec')

            assert profile_arcsec.position == (1.0, 2.0)
            assert profile_arcsec.position[0].unit == 'arcsec'
            assert profile_arcsec.position[1].unit == 'arcsec'
            assert profile_arcsec.param_float == 2.0
            assert profile_arcsec.length == 3.0
            assert profile_arcsec.length.unit == 'arcsec'
            assert profile_arcsec.luminosity == 4.0
            assert profile_arcsec.luminosity.unit == 'eps'
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == 'angular'
            assert profile_arcsec.mass_over_luminosity == 6.0
            assert profile_arcsec.mass_over_luminosity.unit == 'angular / eps'

            profile_kpc = profile_arcsec.new_profile_with_units_converted(unit_length='kpc', kpc_per_arcsec=2.0)

            assert profile_kpc.position == (2.0, 4.0)
            assert profile_kpc.position[0].unit == 'kpc'
            assert profile_kpc.position[1].unit == 'kpc'
            assert profile_kpc.param_float == 2.0
            assert profile_kpc.length == 6.0
            assert profile_kpc.length.unit == 'kpc'
            assert profile_kpc.luminosity == 4.0
            assert profile_kpc.luminosity.unit == 'eps'
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == 'angular'
            assert profile_kpc.mass_over_luminosity == 6.0
            assert profile_kpc.mass_over_luminosity.unit == 'angular / eps'

            profile_kpc = profile_kpc.new_profile_with_units_converted(unit_length='kpc')

            assert profile_kpc.position == (2.0, 4.0)
            assert profile_kpc.position[0].unit == 'kpc'
            assert profile_kpc.position[1].unit == 'kpc'
            assert profile_kpc.param_float == 2.0
            assert profile_kpc.length == 6.0
            assert profile_kpc.length.unit == 'kpc'
            assert profile_kpc.luminosity == 4.0
            assert profile_kpc.luminosity.unit == 'eps'
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == 'angular'
            assert profile_kpc.mass_over_luminosity == 6.0
            assert profile_kpc.mass_over_luminosity.unit == 'angular / eps'

            profile_arcsec = profile_kpc.new_profile_with_units_converted(unit_length='arcsec', kpc_per_arcsec=2.0)

            assert profile_arcsec.position == (1.0, 2.0)
            assert profile_arcsec.position[0].unit == 'arcsec'
            assert profile_arcsec.position[1].unit == 'arcsec'
            assert profile_arcsec.param_float == 2.0
            assert profile_arcsec.length == 3.0
            assert profile_arcsec.length.unit == 'arcsec'
            assert profile_arcsec.luminosity == 4.0
            assert profile_arcsec.luminosity.unit == 'eps'
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == 'angular'
            assert profile_arcsec.mass_over_luminosity == 6.0
            assert profile_arcsec.mass_over_luminosity.unit == 'angular / eps'

        def test__conversion_requires_kpc_per_arcsec_but_does_not_supply_it_raises_error(self):

            profile_arcsec = MockDimensionsProfile(position=(dim.Length(1.0, 'arcsec'), dim.Length(2.0, 'arcsec')),)

            with pytest.raises(exc.UnitsException):
                profile_arcsec.new_profile_with_units_converted(unit_length='kpc')

            profile_kpc = profile_arcsec.new_profile_with_units_converted(unit_length='kpc', kpc_per_arcsec=2.0)

            with pytest.raises(exc.UnitsException):
                profile_kpc.new_profile_with_units_converted(unit_length='arcsec')

        def test__eps_to_counts_conversions_of_luminosity__conversions_convert_values(self):

            profile_eps = MockDimensionsProfile(
                position=(dim.Length(1.0, 'arcsec'), dim.Length(2.0, 'arcsec')),
                param_float=2.0,
                length=dim.Length(value=3.0, unit_length='arcsec'),
                luminosity=dim.Luminosity(value=4.0, unit_luminosity='eps'),
                mass=dim.Mass(value=5.0, unit_mass='angular'),
                mass_over_luminosity=dim.MassOverLuminosity(value=6.0, unit_luminosity='eps', unit_mass='angular'))

            assert profile_eps.position == (1.0, 2.0)
            assert profile_eps.position[0].unit_length == 'arcsec'
            assert profile_eps.position[1].unit_length == 'arcsec'
            assert profile_eps.param_float == 2.0
            assert profile_eps.length == 3.0
            assert profile_eps.length.unit_length == 'arcsec'
            assert profile_eps.luminosity == 4.0
            assert profile_eps.luminosity.unit_luminosity == 'eps'
            assert profile_eps.mass == 5.0
            assert profile_eps.mass.unit_mass == 'angular'
            assert profile_eps.mass_over_luminosity == 6.0
            assert profile_eps.mass_over_luminosity.unit == 'angular / eps'

            profile_eps = profile_eps.new_profile_with_units_converted(unit_luminosity='eps')

            assert profile_eps.position == (1.0, 2.0)
            assert profile_eps.position[0].unit_length == 'arcsec'
            assert profile_eps.position[1].unit_length == 'arcsec'
            assert profile_eps.param_float == 2.0
            assert profile_eps.length == 3.0
            assert profile_eps.length.unit_length == 'arcsec'
            assert profile_eps.luminosity == 4.0
            assert profile_eps.luminosity.unit_luminosity == 'eps'
            assert profile_eps.mass == 5.0
            assert profile_eps.mass.unit_mass == 'angular'
            assert profile_eps.mass_over_luminosity == 6.0
            assert profile_eps.mass_over_luminosity.unit == 'angular / eps'

            profile_counts = profile_eps.new_profile_with_units_converted(unit_luminosity='counts',
                                                                          exposure_time=10.0)

            assert profile_counts.position == (1.0, 2.0)
            assert profile_counts.position[0].unit_length == 'arcsec'
            assert profile_counts.position[1].unit_length == 'arcsec'
            assert profile_counts.param_float == 2.0
            assert profile_counts.length == 3.0
            assert profile_counts.length.unit_length == 'arcsec'
            assert profile_counts.luminosity == 40.0
            assert profile_counts.luminosity.unit_luminosity == 'counts'
            assert profile_counts.mass == 5.0
            assert profile_counts.mass.unit_mass == 'angular'
            assert profile_counts.mass_over_luminosity == pytest.approx(0.6, 1.0e-4)
            assert profile_counts.mass_over_luminosity.unit == 'angular / counts'

            profile_counts = profile_counts.new_profile_with_units_converted(unit_luminosity='counts')


            assert profile_counts.position == (1.0, 2.0)
            assert profile_counts.position[0].unit_length == 'arcsec'
            assert profile_counts.position[1].unit_length == 'arcsec'
            assert profile_counts.param_float == 2.0
            assert profile_counts.length == 3.0
            assert profile_counts.length.unit_length == 'arcsec'
            assert profile_counts.luminosity == 40.0
            assert profile_counts.luminosity.unit_luminosity == 'counts'
            assert profile_counts.mass == 5.0
            assert profile_counts.mass.unit_mass == 'angular'
            assert profile_counts.mass_over_luminosity == pytest.approx(0.6, 1.0e-4)
            assert profile_counts.mass_over_luminosity.unit == 'angular / counts'

            profile_eps = profile_counts.new_profile_with_units_converted(unit_luminosity='eps',
                                                                          exposure_time=10.0)

            assert profile_eps.position == (1.0, 2.0)
            assert profile_eps.position[0].unit_length == 'arcsec'
            assert profile_eps.position[1].unit_length == 'arcsec'
            assert profile_eps.param_float == 2.0
            assert profile_eps.length == 3.0
            assert profile_eps.length.unit_length == 'arcsec'
            assert profile_eps.luminosity == 4.0
            assert profile_eps.luminosity.unit_luminosity == 'eps'
            assert profile_eps.mass == 5.0
            assert profile_eps.mass.unit_mass == 'angular'
            assert profile_eps.mass_over_luminosity == pytest.approx(6.0, 1.0e-4)
            assert profile_eps.mass_over_luminosity.unit == 'angular / eps'

        def test__luminosity_conversion_requires_exposure_time_but_does_not_supply_it_raises_error(self):

            profile_eps = MockDimensionsProfile(
                position=(dim.Length(1.0, 'arcsec'), dim.Length(2.0, 'arcsec')),
                param_float=2.0,
                length=dim.Length(value=3.0, unit_length='arcsec'),
                luminosity=dim.Luminosity(value=4.0, unit_luminosity='eps'),
                mass=dim.Mass(value=5.0, unit_mass='angular'),
                mass_over_luminosity=dim.MassOverLuminosity(value=6.0, unit_luminosity='eps', unit_mass='angular'))

            with pytest.raises(exc.UnitsException):
                profile_eps.new_profile_with_units_converted(unit_luminosity='counts')

            profile_counts = profile_eps.new_profile_with_units_converted(unit_luminosity='counts', exposure_time=10.0)

            with pytest.raises(exc.UnitsException):
                profile_counts.new_profile_with_units_converted(unit_luminosity='eps')

        def test__angular_to_solMass_conversions_of_mass__conversions_convert_values(self):

            profile_angular = MockDimensionsProfile(
                position=(dim.Length(1.0, 'arcsec'), dim.Length(2.0, 'arcsec')),
                param_float=2.0,
                length=dim.Length(value=3.0, unit_length='arcsec'),
                luminosity=dim.Luminosity(value=4.0, unit_luminosity='eps'),
                mass=dim.Mass(value=5.0, unit_mass='angular'),
                mass_over_luminosity=dim.MassOverLuminosity(value=6.0, unit_luminosity='eps', unit_mass='angular'))

            assert profile_angular.position == (1.0, 2.0)
            assert profile_angular.position[0].unit_length == 'arcsec'
            assert profile_angular.position[1].unit_length == 'arcsec'
            assert profile_angular.param_float == 2.0
            assert profile_angular.length == 3.0
            assert profile_angular.length.unit_length == 'arcsec'
            assert profile_angular.luminosity == 4.0
            assert profile_angular.luminosity.unit_luminosity == 'eps'
            assert profile_angular.mass == 5.0
            assert profile_angular.mass.unit_mass == 'angular'
            assert profile_angular.mass_over_luminosity == 6.0
            assert profile_angular.mass_over_luminosity.unit == 'angular / eps'

            profile_angular = profile_angular.new_profile_with_units_converted(unit_mass='angular')

            assert profile_angular.position == (1.0, 2.0)
            assert profile_angular.position[0].unit_length == 'arcsec'
            assert profile_angular.position[1].unit_length == 'arcsec'
            assert profile_angular.param_float == 2.0
            assert profile_angular.length == 3.0
            assert profile_angular.length.unit_length == 'arcsec'
            assert profile_angular.luminosity == 4.0
            assert profile_angular.luminosity.unit_luminosity == 'eps'
            assert profile_angular.mass == 5.0
            assert profile_angular.mass.unit_mass == 'angular'
            assert profile_angular.mass_over_luminosity == 6.0
            assert profile_angular.mass_over_luminosity.unit == 'angular / eps'

            profile_solMass = profile_angular.new_profile_with_units_converted(unit_mass='solMass',
                                                                          critical_surface_density=10.0)

            assert profile_solMass.position == (1.0, 2.0)
            assert profile_solMass.position[0].unit_length == 'arcsec'
            assert profile_solMass.position[1].unit_length == 'arcsec'
            assert profile_solMass.param_float == 2.0
            assert profile_solMass.length == 3.0
            assert profile_solMass.length.unit_length == 'arcsec'
            assert profile_solMass.luminosity == 4.0
            assert profile_solMass.luminosity.unit_luminosity == 'eps'
            assert profile_solMass.mass == 50.0
            assert profile_solMass.mass.unit_mass == 'solMass'
            assert profile_solMass.mass_over_luminosity == pytest.approx(60.0, 1.0e-4)
            assert profile_solMass.mass_over_luminosity.unit == 'solMass / eps'

            profile_solMass = profile_solMass.new_profile_with_units_converted(unit_mass='solMass')

            assert profile_solMass.position == (1.0, 2.0)
            assert profile_solMass.position[0].unit_length == 'arcsec'
            assert profile_solMass.position[1].unit_length == 'arcsec'
            assert profile_solMass.param_float == 2.0
            assert profile_solMass.length == 3.0
            assert profile_solMass.length.unit_length == 'arcsec'
            assert profile_solMass.luminosity == 4.0
            assert profile_solMass.luminosity.unit_luminosity == 'eps'
            assert profile_solMass.mass == 50.0
            assert profile_solMass.mass.unit_mass == 'solMass'
            assert profile_solMass.mass_over_luminosity == pytest.approx(60.0, 1.0e-4)
            assert profile_solMass.mass_over_luminosity.unit == 'solMass / eps'

            profile_angular = profile_solMass.new_profile_with_units_converted(unit_mass='angular',
                                                                          critical_surface_density=10.0)

            assert profile_angular.position == (1.0, 2.0)
            assert profile_angular.position[0].unit_length == 'arcsec'
            assert profile_angular.position[1].unit_length == 'arcsec'
            assert profile_angular.param_float == 2.0
            assert profile_angular.length == 3.0
            assert profile_angular.length.unit_length == 'arcsec'
            assert profile_angular.luminosity == 4.0
            assert profile_angular.luminosity.unit_luminosity == 'eps'
            assert profile_angular.mass == 5.0
            assert profile_angular.mass.unit_mass == 'angular'
            assert profile_angular.mass_over_luminosity == pytest.approx(6.0, 1.0e-4)
            assert profile_angular.mass_over_luminosity.unit == 'angular / eps'

        def test__mass_conversion_requires_critical_surface_density_but_does_not_supply_it_raises_error(self):

            profile_angular = MockDimensionsProfile(
                position=(dim.Length(1.0, 'arcsec'), dim.Length(2.0, 'arcsec')),
                param_float=2.0,
                length=dim.Length(value=3.0, unit_length='arcsec'),
                luminosity=dim.Luminosity(value=4.0, unit_luminosity='eps'),
                mass=dim.Mass(value=5.0, unit_mass='angular'),
                mass_over_luminosity=dim.MassOverLuminosity(value=6.0, unit_luminosity='eps', unit_mass='angular'))

            with pytest.raises(exc.UnitsException):
                profile_angular.new_profile_with_units_converted(unit_mass='solMass')

            profile_solMass = profile_angular.new_profile_with_units_converted(unit_mass='solMass', critical_surface_density=10.0)

            with pytest.raises(exc.UnitsException):
                profile_solMass.new_profile_with_units_converted(unit_mass='angular')


class TestUnitCheckConversionWrapper(object):

    def test__profile_length_units_calculations__profile_is_converted_for_calculation_if_different_to_input_units(self):

        profile = MockDimensionsProfile(length=dim.Length(3.0, 'arcsec'))

        cosmo = MockCosmology(kpc_per_arcsec=2.0)

        # length: arcsec -> arcsec, stays 3.0,  length_input: arcsec -> arcsec, stays 1.0

        length_input = dim.Length(1.0, 'arcsec')
        length = profile.unit_length_calc(length_input=length_input, unit_length='arcsec')
        assert length.unit_length == 'arcsec'
        assert length == 4.0

        # length: arcsec -> arcsec, stays 3.0,  length_input  kpc -> arcsec, converts to 1.0 / 2.0 = 0.5

        length_input = dim.Length(1.0, 'kpc')
        length = profile.unit_length_calc(length_input=length_input, unit_length='arcsec',
                                          redshift_profile=0.5, cosmology=cosmo)
        assert length.unit_length == 'arcsec'
        assert length == 3.5

        # length: arcsec -> kpc, converts to 3.0 * 2.0 = 6.0,  length_input  kpc -> kpc, stays 1.0

        length_input = dim.Length(1.0, 'kpc')
        length = profile.unit_length_calc(length_input=length_input, unit_length='kpc',
                                          redshift_profile=0.5, cosmology=cosmo)
        assert length.unit_length == 'kpc'
        assert length == 7.0
        
        profile = MockDimensionsProfile(length=dim.Length(3.0, 'kpc'))

        # length: kpc -> kpc, stays 3.0,  length_input: kpc -> kpc, stays 1.0

        length_input = dim.Length(1.0, 'kpc')
        length = profile.unit_length_calc(length_input=length_input, unit_length='kpc',
                                          redshift_profile=0.5, cosmology=cosmo)
        assert length.unit_length == 'kpc'
        assert length == 4.0

        # length: kpc -> kpc, stays 3.0,  length_input: arcsec -> kpc, convert to 1.0 * 2.0 = 2.0

        length_input = dim.Length(1.0, 'arcsec')
        length = profile.unit_length_calc(length_input=length_input, unit_length='kpc',
                                          redshift_profile=0.5, cosmology=cosmo)
        assert length.unit_length == 'kpc'
        assert length == 5.0

        # length: kpc -> arcsec, converts to 3.0 / 2.0 = 1.5,  length_input: kpc -> kpc, stays 1.0

        length_input = dim.Length(1.0, 'arcsec')
        length = profile.unit_length_calc(length_input=length_input, unit_length='arcsec',
                                          redshift_profile=0.5, cosmology=cosmo)
        assert length.unit_length == 'arcsec'
        assert length == 2.5

    def test__profile_luminosity_units_calculations__profile_is_converted_for_calculation_if_different_to_input_units(self):
        
        profile = MockDimensionsProfile(luminosity=dim.Luminosity(3.0, 'eps'))

        cosmo = MockCosmology()

        # luminosity: eps -> eps, stays 3.0,  luminosity_input: eps -> eps, stays 1.0

        luminosity_input = dim.Luminosity(1.0, 'eps')
        luminosity = profile.unit_luminosity_calc(luminosity_input=luminosity_input, unit_luminosity='eps')
        assert luminosity.unit_luminosity == 'eps'
        assert luminosity == 4.0

        # luminosity: eps -> eps, stays 3.0,  luminosity_input  counts -> eps, converts to 1.0 / 2.0 = 0.5

        luminosity_input = dim.Luminosity(1.0, 'counts')
        luminosity = profile.unit_luminosity_calc(luminosity_input=luminosity_input, unit_luminosity='eps',
                                                  redshift_profile=0.5, cosmology=cosmo, exposure_time=2.0)
        assert luminosity.unit_luminosity == 'eps'
        assert luminosity == 3.5

        # luminosity: eps -> counts, converts to 3.0 * 2.0 = 6.0,  luminosity_input  counts -> counts, stays 1.0

        luminosity_input = dim.Luminosity(1.0, 'counts')
        luminosity = profile.unit_luminosity_calc(luminosity_input=luminosity_input, unit_luminosity='counts',
                                          redshift_profile=0.5, cosmology=cosmo, exposure_time=2.0)
        assert luminosity.unit_luminosity == 'counts'
        assert luminosity == 7.0

        profile = MockDimensionsProfile(luminosity=dim.Luminosity(3.0, 'counts'))

        # luminosity: counts -> counts, stays 3.0,  luminosity_input: counts -> counts, stays 1.0

        luminosity_input = dim.Luminosity(1.0, 'counts')
        luminosity = profile.unit_luminosity_calc(luminosity_input=luminosity_input, unit_luminosity='counts',
                                          redshift_profile=0.5, cosmology=cosmo, exposure_time=2.0)
        assert luminosity.unit_luminosity == 'counts'
        assert luminosity == 4.0

        # luminosity: counts -> counts, stays 3.0,  luminosity_input: eps -> counts, convert to 1.0 * 2.0 = 2.0

        luminosity_input = dim.Luminosity(1.0, 'eps')
        luminosity = profile.unit_luminosity_calc(luminosity_input=luminosity_input, unit_luminosity='counts',
                                          redshift_profile=0.5, cosmology=cosmo, exposure_time=2.0)
        assert luminosity.unit_luminosity == 'counts'
        assert luminosity == 5.0

        # luminosity: counts -> eps, converts to 3.0 / 2.0 = 1.5,  luminosity_input: counts -> counts, stays 1.0

        luminosity_input = dim.Luminosity(1.0, 'eps')
        luminosity = profile.unit_luminosity_calc(luminosity_input=luminosity_input, unit_luminosity='eps',
                                          redshift_profile=0.5, cosmology=cosmo, exposure_time=2.0)
        assert luminosity.unit_luminosity == 'eps'
        assert luminosity == 2.5

    def test__profile_mass_units_calculations__profile_is_converted_for_calculation_if_different_to_input_units(self):

        profile = MockDimensionsProfile(length=dim.Length(1.0, 'arcsec'), mass=dim.Mass(3.0, 'angular'))

        cosmo = MockCosmology(critical_surface_density=2.0)

        # mass: angular -> angular, stays 3.0,  mass_input: angular -> angular, stays 1.0

        mass_input = dim.Mass(1.0, 'angular')
        mass = profile.unit_mass_calc(mass_input=mass_input, unit_mass='angular',
                                      redshift_profile=0.5, redshift_source=1.0, cosmology=cosmo)
        assert mass.unit_mass == 'angular'
        assert mass == 4.0

        # mass: angular -> angular, stays 3.0,  mass_input  solMass -> angular, stays as 1.0

        mass_input = dim.Mass(1.0, 'solMass')
        mass = profile.unit_mass_calc(mass_input=mass_input, unit_mass='angular',
                                      redshift_profile=0.5, redshift_source=1.0, cosmology=cosmo)
        assert mass.unit_mass == 'angular'
        assert mass == 4.0

        # mass: angular -> solMass, converts to 3.0 * 2.0 = 6.0,  mass_input  solMass -> solMass, stays 1.0

        mass_input = dim.Mass(1.0, 'solMass')
        mass = profile.unit_mass_calc(mass_input=mass_input, unit_mass='solMass',
                                          redshift_profile=0.5, redshift_source=1.0, cosmology=cosmo)
        assert mass.unit_mass == 'solMass'
        assert mass == 7.0

        profile = MockDimensionsProfile(length=dim.Length(1.0, 'arcsec'), mass=dim.Mass(3.0, 'solMass'))

        # mass: solMass -> solMass, stays 3.0,  mass_input: solMass -> solMass, stays 1.0

        mass_input = dim.Mass(1.0, 'solMass')
        mass = profile.unit_mass_calc(mass_input=mass_input, unit_mass='solMass',
                                          redshift_profile=0.5, redshift_source=1.0, cosmology=cosmo)
        assert mass.unit_mass == 'solMass'
        assert mass == 4.0

        # mass: solMass -> solMass, stays 3.0,  mass_input: angular -> solMass, convert to 1.0 * 2.0 = 2.0

        mass_input = dim.Mass(1.0, 'angular')
        mass = profile.unit_mass_calc(mass_input=mass_input, unit_mass='solMass',
                                          redshift_profile=0.5, redshift_source=1.0, cosmology=cosmo)
        assert mass.unit_mass == 'solMass'
        assert mass == 5.0

        # mass: solMass -> angular, stays 3.0,  mass_input: solMass -> solMass, stays 1.0

        mass_input = dim.Mass(1.0, 'angular')
        mass = profile.unit_mass_calc(mass_input=mass_input, unit_mass='angular',
                                          redshift_profile=0.5, redshift_source=1.0, cosmology=cosmo)
        assert mass.unit_mass == 'angular'
        assert mass == 4.0


from astropy import cosmology
from astropy import constants
from astropy import units
from autolens.model.profiles import mass_profiles as mp

import numpy as np

cosmo = cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)

z_nfw = 0.6
z_source = 2.5

nfw = mp.SphericalTruncatedNFWChallenge(kappa_s=0.08, scale_radius=0.3)

concentration = nfw.concentration_for_units(redshift_profile=z_nfw, redshift_source=z_source, unit_length='kpc',
                                     unit_mass='solMass', redshift_of_cosmic_average_density='profile',
                                     cosmology=cosmo)

mass_at_200 = nfw.mass_at_200_for_units(redshift_profile=z_nfw, redshift_source=z_source, unit_length='kpc',
                                     unit_mass='solMass', redshift_of_cosmic_average_density='profile',
                                     cosmology=cosmo)

print('Concentration = ', concentration)
print('Mass at 200 = {:.4e}'.format(mass_at_200))

nfw = mp.SphericalTruncatedNFWChallenge(kappa_s=0.01, scale_radius=0.07)

concentration = nfw.concentration_for_units(redshift_profile=z_nfw, redshift_source=z_source, unit_length='kpc',
                                     unit_mass='solMass', redshift_of_cosmic_average_density='profile',
                                     cosmology=cosmo)

mass_at_200 = nfw.mass_at_200_for_units(redshift_profile=z_nfw, redshift_source=z_source, unit_length='kpc',
                                     unit_mass='solMass', redshift_of_cosmic_average_density='profile',
                                     cosmology=cosmo)

print('Concentration = ', concentration)
print('Mass at 200 = {:.4e}'.format(mass_at_200))


nfw = mp.SphericalTruncatedNFWChallenge(kappa_s=0.002, scale_radius=0.02)

concentration = nfw.concentration_for_units(redshift_profile=z_nfw, redshift_source=z_source, unit_length='kpc',
                                     unit_mass='solMass', redshift_of_cosmic_average_density='profile',
                                     cosmology=cosmo)

mass_at_200 = nfw.mass_at_200_for_units(redshift_profile=z_nfw, redshift_source=z_source, unit_length='kpc',
                                     unit_mass='solMass', redshift_of_cosmic_average_density='profile',
                                     cosmology=cosmo)

print('Concentration = ', concentration)
print('Mass at 200 = {:.4e}'.format(mass_at_200))


from astropy import cosmology
from scipy.optimize import fsolve
from autolens.model.profiles import mass_profiles as mp
from astropy import constants
import numpy as np

from autolens.model import cosmology_util

cosmo=cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)

z_lens = 0.6
z_source = 2.5

C_v = constants.c.to('km / s')
G_crit = 4.302e-6  # kpc M_sun^-1 (km/s)^2
G_crit = constants.G.to( 'kpc km2 / (solMass s2)')

s_crit = (C_v * C_v * cosmo.angular_diameter_distance(z_source) / (4. * np.pi * cosmo.angular_diameter_distance(z_lens) *
           cosmo.angular_diameter_distance_z1z2(z_lens, z_source)) /
          (1000.0 * G_crit))  # critical density at los halo

const = (C_v * C_v) / (4 * np.pi) / (1000.0 * G_crit)

print(const)

angs = (cosmo.angular_diameter_distance(z_source) / (cosmo.angular_diameter_distance(z_lens) *cosmo.angular_diameter_distance_z1z2(z_lens, z_source)))

s_crit_al = cosmology_util.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
    redshift_0=z_lens, redshift_1=z_source, cosmology=cosmo, unit_length='kpc', unit_mass='solMass')

print(s_crit)
print(s_crit_al)